
'''
This file is automatically generated; Do not edit it. :)
'''
VERSION_INFO = {
    'final': True,
    'version': '1.27.3',
    'branch_nick': 'release-1.27',
    'revision_id': '5d0133b7b33f9a8854c0db45ed61cb54a1c0a33e',
    'revno': 11181
}
